@@ -76,7 +76,7 @@ struct _GstControlPoint
     } cubic;
     struct { /* 24 bytes */
       gdouble c1s, c2s, c3s;
-    } cubic_mono;
+    } cubic_monotonic;
     guint8 _gst_reserved[64];
   } cache;
 };
@@ -120,7 +120,7 @@ GType gst_timed_value_control_source_get_type (void);
 /* Functions */
 
 GSequenceIter * gst_timed_value_control_source_find_control_point_iter (
-                                                               GstTimedValueControlSource * self, 
+                                                               GstTimedValueControlSource * self,
                                                                GstClockTime timestamp);
 
 gboolean        gst_timed_value_control_source_set            (GstTimedValueControlSource * self,
